/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=V
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=mercurio 6 marte 4
*/

/* RECORD
Alberto Hernández Chillón
students of the University of Murcia, Spain
time 100 msec
speed-up 4
*/ 

/*Blocks of rows are assigned to the processes,
and an additional matrix is used to store the new values at
each step.
OpenMP threads are not used, and their use could give additional reduction of the time.
6 processes are started in mercurio and 4 in marte because some processes where running in
marte, and this combination gave the best result.
The best option should be 6 + 6 processes.*/

#include <stdlib.h>
#include <omp.h>
#include <string.h>
#include <mpi.h>

void sec(unsigned char *Imagen, unsigned int Size, int Rep, int nodo, int np)
{
	int k, cellFAnterior, cellActual, cellFSiguiente, aux, filas, start = 0;
	unsigned int stride, i, j;
	unsigned char *auxImagen;
	MPI_Status estado;

	/*Broadcast del tamaño y las repeticiones a realizar...*/
	MPI_Bcast(&Size, 1, MPI_UNSIGNED, 0, MPI_COMM_WORLD);
	MPI_Bcast(&Rep, 1, MPI_INT, 0, MPI_COMM_WORLD);

	/*Calculo de las filas sobre las que debe operar cada nodo, luego deberemos sumar las dos de guia*/
	filas = Size/(np);//Parte entera
	if(Size%np > nodo) filas++;//Parte decimal

	stride = Size+2;
	auxImagen = (unsigned char *)calloc(stride*filas + stride*2, sizeof(unsigned char));

	if(nodo==0)
        {
		start = start + filas;
		aux = filas;	/*Guardamos temporalmente las filas del nodo 0...*/
                for(i=1;i<np;i++)
                {
			filas=Size/np;//Parte entera
			if(Size%np > i) filas++;//Parte decimal
			MPI_Send(&Imagen[start*stride], filas*stride + stride*2, MPI_UNSIGNED_CHAR, i, 17, MPI_COMM_WORLD);
                        start = start + filas;
			/*Envio a cada nodo desde el maestro sus filas correspondientes*/
                }
		filas = aux;	/*Restauramos el valor*/
        }
	else
	{
		Imagen = (unsigned char *)calloc(stride*filas + stride*2, sizeof(unsigned char));
                MPI_Recv(Imagen, filas*stride + stride*2, MPI_UNSIGNED_CHAR, 0, 17, MPI_COMM_WORLD, &estado);
	}
	/*Cada nodo realiza las operaciones entre las dos matrices como en la version con OpenMP, excepto por la sincronizacion final*/
	for (k=1; k <= Rep; k++)
	{
		if(k%2 != 0)
		{
			for(i=0; i < filas+2; i++)
			{
				for(j=0; j < Size+2; j++)
				{
					cellActual=(i*stride) + j;
					cellFAnterior=cellActual - stride;
					cellFSiguiente=cellActual + stride;

					if(i == 0 || i == filas+1 || j == 0 || j == Size+1)auxImagen[cellActual] = Imagen[cellActual];
					else
					auxImagen[cellActual]=(Imagen[cellFAnterior-1] + Imagen[cellFAnterior] + Imagen[cellFAnterior+1] +
							Imagen[cellActual-1] + Imagen[cellActual] + Imagen[cellActual+1] +
							Imagen[cellFSiguiente-1] + Imagen[cellFSiguiente] + Imagen[cellFSiguiente+1]) / 9;

				}
			}
		}
		else
		{
			for(i=0; i < filas+2; i++)
			{
				for(j=0; j < Size+2; j++)
				{
					cellActual=(i*stride) + j;
					cellFAnterior=cellActual - stride;
					cellFSiguiente=cellActual + stride;

					if(i == 0 || i == filas+1 || j == 0 || j == Size+1)Imagen[cellActual] = auxImagen[cellActual];
					else
					Imagen[cellActual]=(auxImagen[cellFAnterior-1] + auxImagen[cellFAnterior] + auxImagen[cellFAnterior+1] +
							auxImagen[cellActual-1] + auxImagen[cellActual] + auxImagen[cellActual+1] +
							auxImagen[cellFSiguiente-1] + auxImagen[cellFSiguiente] + auxImagen[cellFSiguiente+1]) / 9;

				}
			}

		}
/*
* Sincronizacion.
* Reparto de informacion: Cada nodo n con 1 < n < np-1 pasa su segunda fila al nodo n-1, y su penultima fila al n+1.
* Cada nodo n=0 pasa su penultima fila al nodo 1.
* Cada nodo n=np-1 pasa su segunda fila al nodo np-2.
* Para una iteracion no es necesario sincronizar los nodos, para la ultima tampoco.
*/
		if(k != Rep){
			if(nodo == 0)
			{
				MPI_Send(k%2 != 0? &auxImagen[filas*stride] : &Imagen[filas*stride],
					stride, MPI_UNSIGNED_CHAR, nodo+1, 77, MPI_COMM_WORLD);
				MPI_Recv(k%2 != 0? &auxImagen[filas*stride+stride] : &Imagen[filas*stride+stride],
					stride, MPI_UNSIGNED_CHAR, nodo+1, 77, MPI_COMM_WORLD, &estado);
			}else if(nodo != np-1)
			{

				MPI_Send(k%2 != 0? &auxImagen[stride] : &Imagen[stride],
					stride, MPI_UNSIGNED_CHAR, nodo-1, 77, MPI_COMM_WORLD);

				MPI_Send(k%2 != 0? &auxImagen[filas*stride] : &Imagen[filas*stride],
					stride, MPI_UNSIGNED_CHAR, nodo+1, 77, MPI_COMM_WORLD);

				MPI_Recv(k%2 != 0? &auxImagen[0] : &Imagen[0],
					stride, MPI_UNSIGNED_CHAR, nodo-1, 77, MPI_COMM_WORLD, &estado);

				MPI_Recv(k%2 != 0? &auxImagen[filas*stride+stride] : &Imagen[filas*stride+stride],
					stride, MPI_UNSIGNED_CHAR, nodo+1, 77, MPI_COMM_WORLD, &estado);

			}else
			{
				MPI_Send(k%2 != 0? &auxImagen[stride] : &Imagen[stride],
					stride, MPI_UNSIGNED_CHAR, nodo-1, 77, MPI_COMM_WORLD);
				MPI_Recv(k%2 != 0? &auxImagen[0] : &Imagen[0],
					stride, MPI_UNSIGNED_CHAR, nodo-1, 77, MPI_COMM_WORLD, &estado);
			}
		}
	}

	if(Rep%2 != 0)memcpy(Imagen, auxImagen, stride*filas + stride*2);
	free(auxImagen);

	/*Calculos hechos. Uniendo resultados...*/
	if(nodo == 0)
	{
		start = filas;
		aux = filas;
                for(i=1;i<np;i++)
                {
                	filas=Size/np;//Parte entera
			if(Size%np > i) filas++;//Parte decimal
			MPI_Recv(&Imagen[start*stride + stride], filas*stride, MPI_UNSIGNED_CHAR, i, 71, MPI_COMM_WORLD, &estado);
                        start = start + filas;
                }
		filas = aux;
	}
	else
	{	/*NO se envian las filas de guia, como es normal...por eso no sincronizamos*/
		MPI_Send(&Imagen[stride], filas*stride, MPI_UNSIGNED_CHAR, 0, 71, MPI_COMM_WORLD);
		free(Imagen);
	}
}
